/* Jay Clary [jec]                                                        */
/* Implementation: Borland C++, version 2.0                               */
/* Compiled in COMPACT memory model                                       */
/*                                                                        */
/* This is a program written to explore RLE packing of color values using */
/* assembler to speed-up the process. This is my first attempt at writing */
/* in assembler to do anything truly useful. I didn't think I knew enough */
/* about assembler to make it work, but was pleasantly surprised at how   */
/* quickly the pieces fell into place. I'm sure there are some places in  */
/* the code that could be better optimized and I welcome any suggestions. */
/* If you have any use for this code, feel free to incorporate it into    */
/* your application.                                                      */


#include <stdio.h>
#include <stdlib.h>
#include <alloc.h>

#define Cnt 8192
#define UBYTE unsigned char
#define UWORD unsigned int


   UBYTE far *pa, far *p, far *qa;
   UWORD j, NumRuns;

   UWORD RLEPack( UBYTE far *, UBYTE far *, UWORD );

void main()

   {

      printf("Assembler RLF Packing Test\n\n");

      // pa is the input data stream

      if ((pa = (UBYTE far *) farmalloc(Cnt)) == NULL) {

         printf("Not enough memory for input\n");
         exit(99);

      }

      // qa will be the output data stream

      if ((qa = (UBYTE far *) farmalloc(Cnt*2)) == NULL) {

         printf("Not enough memory for output\n");
         free(pa);
         exit(99);

      }


      // Set some random data in the input to test RLE

      for (j=0, p=pa; j < Cnt; j++, p++) (*p) = random(4);

      NumRuns = RLEPack( pa, qa, Cnt );


      printf("%d bytes packed into %d runs\n\n", Cnt, NumRuns );
      free(pa);
      free(qa);
      return;

   }


   /* Since this RLE packing method is written to support a file format    */
   /* that we have no control over (it is from another software package)   */
   /* we can not alter the RLE encoding. The main problem is that the      */
   /* the RLE method could make the output file twice the size of the      */
   /* input. It should be written intelligently like the RLE packing of    */
   /* PCX image files. In PCX, a special code is used to determine if      */
   /* following byte is encoded or not. This way, you can guarentee that   */
   /* the RLE packing will NEVER be larger than the original, at most it   */
   /* will be the same size. Because of this, the output buffer must be    */
   /* allocated to twice the size of the original to handle the worst-case */
   /* senerio.                                                             */


   UWORD RLEPack( UBYTE far *inbuffer, UBYTE far *outbuffer, UWORD inbuffsize )

      {

         //  Register usage
         //
         //  ds:si = points to the input buffer
         //  es:di = points to the output buffer
         //  cx    = holds input buffer size. Used to drive the loop
         //  bx    = accumulates the total number of runs created
         //  ah    = accumulates current run total. Maximum run count
         //          in one run is 255.
         //  al    = holds current color value


         asm {        push ds                  // save the ds register
                      lds  si, inbuffer        // ds:si = inbuffer;
                      les  di, outbuffer       // es:di = outbuffer;

                      mov  cx, inbuffsize      // prime loop counter
                      xor  bx, bx              // Zero total run cnt
         }

         newrun:

         asm {        mov  al, ds:[si]         // get the next color
                      mov  ah, 1               // set current run count to 1
                      inc  si                  // point to the next color
         }


         loop1:

         asm {        dec  cx                  // dec loop cnt
                      jcxz kickout             // if (cx==0) exit & cleanup

                      cmp  ah, 0xff            // if (run is full)
                      je   storerun            //    place run in outbuffer

                      cmp  al, ds:[si]         // if (last color != cur color)
                      jne  storerun            //    place run in outbuffer

                      inc  ah                  // else inc current run cnt
                      inc  si                  //      inc input ptr
                      jmp  loop1               //      jump to loop1
         }

         storerun:

         asm {        mov  es:[di], ax         // put run & col in outbuffer
                      inc  di                  // move output ptr to next byte
                      inc  di                  // move one more byte
                      inc  bx                  // inc total runs cnt
                      jmp  newrun              // setup for next run

         }


         kickout:

         asm {        mov  es:[di], ax         // clear the current run
                      inc  bx                  // inc total runs cnt
                      pop  ds                  // restore the ds

         }

         return( _BX );                        // return the total run count
                                               // so you know how many bytes
                                               // to write to disk
      }
